from pathlib import Path

import pytest
import rdflib
from typer.testing import CliRunner

from kurra.db.gsp import clear, delete, exists, get, post, put, upload
from kurra.sparql import query
from kurra.utils import load_graph

runner = CliRunner()

TESTS_DIR = Path(__file__).resolve().parent
LANG_TEST_VOC = Path(__file__).parent.parent / "sparql" / "language-test.ttl"
THREE_TRIPLE_FILE = Path(__file__).parent.parent / "file" / "prefixes-test.ttl"
TESTING_GRAPH = "https://example.com/testing-graph"


def test_exists(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    g1 = exists(sparql_endpoint, "http://nothing.com", http_client)
    assert not g1

    upload(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, http_client)
    g2 = exists(sparql_endpoint, TESTING_GRAPH, http_client)
    assert g2

    delete(sparql_endpoint, TESTING_GRAPH, http_client)
    g3 = exists(sparql_endpoint, TESTING_GRAPH, http_client)
    assert not g3


def test_get(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    g_result = load_graph(LANG_TEST_VOC)

    upload(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, http_client)
    g = get(sparql_endpoint, TESTING_GRAPH, http_client=http_client)
    assert g.isomorphic(g_result)
    assert str(g.identifier) == TESTING_GRAPH

    d = """
        PREFIX : <http://example.com/>
        
        :a 
            :b :c ;
            :d :e ;
        .
        """
    upload(sparql_endpoint, d, None, http_client)

    g2 = get(sparql_endpoint, "default", http_client=http_client)
    assert len(g2) == 2
    assert isinstance(g2.identifier, rdflib.BNode)

    g3 = get(sparql_endpoint, "http://nothing.com", http_client=http_client)
    assert g3 == 404

    g4 = get(
        sparql_endpoint, "default", http_client=http_client, return_format="original"
    )
    assert "http://example.com/" in g4


def test_put(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    put(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, http_client=http_client)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?c) AS ?count) WHERE {?c a skos:Concept}",
        namespaces={"skos": "http://www.w3.org/2004/02/skos/core#"},
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 7

    # replace that graph
    put(sparql_endpoint, THREE_TRIPLE_FILE, TESTING_GRAPH, http_client=http_client)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?s) AS ?count) WHERE { GRAPH <"
        + TESTING_GRAPH
        + "> {?s ?p ?o}}",
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 3


def test_post(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    post(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, http_client=http_client)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?c) AS ?count) WHERE {?c a skos:Concept}",
        namespaces={"skos": "http://www.w3.org/2004/02/skos/core#"},
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 7

    # add to that graph
    post(sparql_endpoint, THREE_TRIPLE_FILE, TESTING_GRAPH, http_client=http_client)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?s) AS ?count) WHERE { GRAPH <"
        + TESTING_GRAPH
        + "> {?s ?p ?o}}",
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 80


def test_delete(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    put(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, http_client=http_client)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?c) AS ?count) WHERE {?c a skos:Concept}",
        namespaces={"skos": "http://www.w3.org/2004/02/skos/core#"},
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 7

    delete(sparql_endpoint, TESTING_GRAPH, http_client=http_client)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?s) AS ?count) WHERE { GRAPH <"
        + TESTING_GRAPH
        + "> {?s ?p ?o}}",
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 0

    assert not exists(sparql_endpoint, TESTING_GRAPH)


def test_clear():
    pass  # alias of delete


def test_upload(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?s) AS ?count) WHERE { GRAPH <"
        + TESTING_GRAPH
        + "> {?s ?p ?o}}",
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 77

    upload(sparql_endpoint, THREE_TRIPLE_FILE, TESTING_GRAPH)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?s) AS ?count) WHERE { GRAPH <"
        + TESTING_GRAPH
        + "> {?s ?p ?o}}",
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 3

    upload(sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, append=True)

    r = query(
        sparql_endpoint,
        "SELECT (COUNT(?s) AS ?count) WHERE { GRAPH <"
        + TESTING_GRAPH
        + "> {?s ?p ?o}}",
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )
    assert r[0]["count"] == 80


@pytest.mark.skip(
    reason="Test works with normal Fuseki but not testing container version"
)
def test_upload_no_graph(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    q = """
        SELECT (COUNT(?s) AS ?c)
        WHERE {
        GRAPH ?g {
                ?s ?p ?o
            }
        }
        """
    r = query(
        sparql_endpoint,
        q,
        return_format="python",
        return_bindings_only=True,
        http_client=http_client,
    )

    assert r[0]["c"] == 77


def test_upload_url(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        sparql_endpoint,
        "https://raw.githubusercontent.com/Kurrawong/kurra/refs/heads/main/tests/db/config.ttl",
        TESTING_GRAPH,
    )

    q = """
        SELECT (COUNT(?s) AS ?c)
        WHERE {
            GRAPH ?g {
                ?s ?p ?o
            }
        }
        """
    r = query(sparql_endpoint, q, return_format="python", return_bindings_only=True)

    assert r[0]["c"] == 142

    # now test one with Content Negotiation and a redirect
    clear(sparql_endpoint, TESTING_GRAPH, http_client)

    upload(sparql_endpoint, "https://linked.data.gov.au/def/vocdermods", TESTING_GRAPH)

    r = query(sparql_endpoint, q, return_format="python", return_bindings_only=True)

    assert r[0]["c"] == 86



import time
from pathlib import Path

import docker
import httpx
import pytest
from testcontainers.core.container import DockerContainer

FUSEKI_IMAGE = "ghcr.io/kurrawong/fuseki-geosparql:git-main-e642d849"


def wait_for_logs(container, text, timeout=30, interval=0.5):
    """
    Wait until the container emits a log line containing `text`.
    """
    client = docker.from_env()
    start_time = time.time()

    logs_seen = ""

    while True:
        # Read logs incrementally
        logs = client.containers.get(container._container.id).logs().decode("utf-8")
        if text in logs:
            return True

        if time.time() - start_time > timeout:
            raise TimeoutError(f"Timed out waiting for log: {text}")

        time.sleep(interval)


@pytest.fixture(scope="function")
def fuseki_container(request: pytest.FixtureRequest):
    container = DockerContainer(FUSEKI_IMAGE)
    container.with_volume_mapping(
        str(Path(__file__).parent.parent / "db" / "shiro.ini"),
        "/fuseki/shiro.ini",
    )
    container.with_volume_mapping(
        str(Path(__file__).parent.parent / "db" / "config.ttl"),
        "/fuseki/config.ttl",
    )
    container.with_exposed_ports(3030)
    container.start()
    wait_for_logs(container, "Started")

    def cleanup():
        container.stop()

    request.addfinalizer(cleanup)
    return container


@pytest.fixture(scope="function")
def http_client(request: pytest.FixtureRequest):
    _http_client = httpx.Client(auth=("admin", "admin"))

    def cleanup():
        _http_client.close()

    request.addfinalizer(cleanup)
    return _http_client



# Olis Graph Functions

from datetime import datetime
from pathlib import Path

import httpx
from rdflib import Dataset, Graph, Literal, URIRef
from rdflib.namespace import RDF, SDO, XSD, DefinedNamespace, Namespace
from rdflib.plugins.parsers.notation3 import BadSyntax

import kurra.db.gsp
from kurra.shacl import validate
from kurra.utils import get_system_graph, load_graph, put_system_graph

OLIS = Namespace("https://olis.dev/")
SYSTEM_GRAPH_IRI = URIRef("https://olis.dev/SystemGraph")
BACKGROUND_GRAPH_IRI = URIRef("http://background")


class OLIS(DefinedNamespace):
    """The namespace for the Olis ontology"""

    _NS = Namespace("https://olis.dev/")
    _fail = True

    RealGraph: URIRef
    VirtualGraph: URIRef

    includes: URIRef
    hasBaseGRaph: URIRef
    hasGraphRole: URIRef
    hasRulesGraph: URIRef


class OLIS_GRAPH_ROLES(DefinedNamespace):
    """The namespace for the Olis Graph Roles vocabulary"""

    _NS = Namespace("http://olis.dev/GraphRoles/")
    _fail = True

    Original: URIRef
    Inferred: URIRef
    Added: URIRef
    Removed: URIRef
    Rules: URIRef


def include(
    including_graph_iri: URIRef | str,
    graphs_to_include_iris: list[URIRef | str],
    system_graph_source: str | Path | Dataset | Graph = None,
    include_background_graph: bool = True,
    http_client: httpx.Client | None = None,
) -> Graph | None:
    """Creates a Virtual Graph with IRI of including_graph_iri that contains all the graphs given in graphs_to_include_iris.

    Args:
        including_graph_iri: the IRI of the graph that will include the other graphs
        graphs_to_include_iris: the IRIs of each graph to be included in the including graph
        system_graph_source: the SPARQL Endpoint, file (RDFLib Graph or Dataset) or an RDFLib Graph or Dataset object or a Path to a Graph or Dataset object to both read and write System Graph info to. If None, a new System Graph object will be returned
        include_background_graph: whether to include http://background in the subsumption
        http_client: an optional HTTPX Client to contain credentials if needed to access the SPARQL Endpoint

    Returns:
        If a system_graph_target is given, the updates System Graph information will be written to it or a
        new System Graph, RDFLib Graph object will be returned

    """
    # value check inputs
    if including_graph_iri is None:
        raise ValueError(
            "You must specify an including_graph_iri as a string or a URIRef"
        )

    if graphs_to_include_iris is None:
        raise ValueError(
            "You must specify at least one string or URIRef in the list variable graphs_to_include_iris"
        )

    if isinstance(including_graph_iri, str):
        including_graph_iri = URIRef(including_graph_iri)

    if isinstance(graphs_to_include_iris[0], str):
        graphs_to_include_iris = [URIRef(x) for x in graphs_to_include_iris]

    if including_graph_iri in graphs_to_include_iris:
        raise ValueError(
            "You cannot include the including_graph_iri in the list of graphs to include"
        )

    if including_graph_iri == BACKGROUND_GRAPH_IRI:
        raise ValueError("You cannot use the background graph to include other graphs")

    if (
        including_graph_iri == SYSTEM_GRAPH_IRI
        or SYSTEM_GRAPH_IRI in graphs_to_include_iris
    ):
        raise ValueError(
            "You cannot use the system graph to include other graphs or include it in other graphs"
        )

    # get/make the system graph
    """
    To extend http://graph-a to include graph-a content and http://graph-b content:
    
    include_graph("http://graph-a", ["http://graph-b"])
    
    This will:
    
    * learn that http://graph-a is a Virtual Graph - since it includes things
    * move any existing http://graph-a content, if it's a Real Graph, to a new Real Graph - http://graph-a-real - by convention
    * state that http://graph-a includes http://graph-a-real and all graphs in the graphs_to_include_iris list 
    """
    system_graph = get_system_graph(system_graph_source, http_client=http_client)

    # alter the system graph as per the inputs
    """
    if including_graph_iri in system_graph:
        if it is an RG:
            pop out RG content and include
    else:
        create it
        
    include all graphs_to_include_iris
    """
    if (including_graph_iri, RDF.type, OLIS.RealGraph) in system_graph:
        # swap out all triples targeting the RG to the new RG
        new_including_graph_real_iri = URIRef(str(including_graph_iri) + "-real")

        for p, o in system_graph.subject_objects(including_graph_iri):
            system_graph.remove((including_graph_iri, p, o))
            system_graph.add((new_including_graph_real_iri, p, o))

        for s, p in system_graph.subject_predicates(including_graph_iri):
            system_graph.remove((s, including_graph_iri, p))
            system_graph.add((s, new_including_graph_real_iri, p))

        # create the new VG
        system_graph.add((including_graph_iri, RDF.type, OLIS.VirtualGraph))

        # include the new RG in the VG
        system_graph.add(
            (including_graph_iri, OLIS.includes, new_including_graph_real_iri)
        )
        system_graph.add((new_including_graph_real_iri, RDF.type, OLIS.RealGraph))
        system_graph.add(
            (
                new_including_graph_real_iri,
                SDO.description,
                Literal(f"Graph containing triples from <{str(including_graph_iri)}>"),
            )
        )
    elif (including_graph_iri, RDF.type, OLIS.VirtualGraph) in system_graph:
        pass  # we will add to the VG below
    else:  # the including_graph_iri is not present in the system_graph so create it
        system_graph.add((including_graph_iri, RDF.type, OLIS.VirtualGraph))
        system_graph.add(
            (
                including_graph_iri,
                SDO.dateCreated,
                Literal(datetime.now().isoformat()[:19], datatype=XSD.dateTime),
            )
        )

    # in all cases now...
    for graph_to_include_iri in graphs_to_include_iris:
        system_graph.add((including_graph_iri, OLIS.includes, graph_to_include_iri))

    system_graph.remove((including_graph_iri, SDO.dateModified, None))
    system_graph.add(
        (
            including_graph_iri,
            SDO.dateModified,
            Literal(datetime.now().isoformat()[:19], datatype=XSD.dateTime),
        )
    )

    if include_background_graph:
        system_graph.add((including_graph_iri, OLIS.includes, BACKGROUND_GRAPH_IRI))

    # return or write altered system graph, depending on system graph source
    return put_system_graph(system_graph, system_graph_source, http_client=http_client)


def exclude(
    including_graph_iri: URIRef | str,
    graphs_to_exclude_iris: list[URIRef | str],
    system_graph_source: str | Path | Dataset | Graph = None,
    http_client: httpx.Client | None = None,
):
    """
    Removed the includes statement for a Virtual Graph to a Named Graph.

    See include() for parameter details
    """
    # value check inputs
    if including_graph_iri is None:
        raise ValueError(
            "You must specify an including_graph_iri as a string or a URIRef"
        )

    if graphs_to_exclude_iris is None:
        raise ValueError(
            "You must specify at least one string or URIRef in the list variable graphs_to_include_iris"
        )

    if isinstance(including_graph_iri, str):
        including_graph_iri = URIRef(including_graph_iri)

    if isinstance(graphs_to_exclude_iris[0], str):
        graphs_to_exclude_iris = [URIRef(x) for x in graphs_to_exclude_iris]

    if including_graph_iri in graphs_to_exclude_iris:
        raise ValueError(
            "You cannot include the including_graph_iri in the list of graphs to exclude"
        )

    if including_graph_iri == BACKGROUND_GRAPH_IRI:
        raise ValueError("You cannot exclude graphs from the background graph")

    if (
        including_graph_iri == SYSTEM_GRAPH_IRI
        or SYSTEM_GRAPH_IRI in graphs_to_exclude_iris
    ):
        raise ValueError(
            "You cannot exclude graph from the system graph or include it in other graphs"
        )

    # get/make the system graph
    system_graph = get_system_graph(system_graph_source, http_client=http_client)

    # alter the system graph as per the inputs
    if not (including_graph_iri, RDF.type, OLIS.VirtualGraph) in system_graph:
        return ValueError(
            "The including_graph_iri is not known as a VirtualGraph in the system_graph_source"
        )

    system_graph.remove((including_graph_iri, SDO.dateModified, None))
    system_graph.add(
        (
            including_graph_iri,
            SDO.dateModified,
            Literal(datetime.now().isoformat()[:19], datatype=XSD.dateTime),
        )
    )

    for graph_to_exclude_iri in graphs_to_exclude_iris:
        system_graph.remove((including_graph_iri, OLIS.includes, graph_to_exclude_iri))

    # return or write altered system graph, depending on system graph source
    return put_system_graph(system_graph, system_graph_source, http_client=http_client)


def validate_system_graph(
    system_graph_source: str | Path | Dataset | Graph, http_client: httpx.Client = None
):
    """Validates a System Graph using the Olis System Graph validator"""
    system_graph = get_system_graph(system_graph_source, http_client=http_client)
    v = validate(system_graph, "https://olis.dev/sg-validator")
    return v



from kurra.sparql import query


def test_gswa():
    # this ednpoint doesn't allow POST and replies with 405
    SPARQL_ENDPOINT = "https://api.vocabulary.gswa.kurrawong.ai/sparql"

    q = """
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    SELECT *
    WHERE {
        ?s a skos:Concept
    } LIMIT 10
    """

    r = query(SPARQL_ENDPOINT, q, return_format="python", return_bindings_only=True)
    assert len(r) == 10


def test_idn():
    # this ednpoint doesn't allow POST and replies with 422
    SPARQL_ENDPOINT = "https://api.idnau.org/sparql"

    q = """
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    SELECT *
    WHERE {
        ?s a skos:Concept
    } LIMIT 10
    """

    r = query(SPARQL_ENDPOINT, q, return_format="python", return_bindings_only=True)
    assert len(r) == 10



from pathlib import Path
from textwrap import dedent

from typer.testing import CliRunner

from kurra.cli import app
from kurra.db.gsp import upload

runner = CliRunner()

LANG_TEST_VOC = Path(__file__).parent.parent.resolve() / "sparql" / "language-test.ttl"
TESTING_GRAPH = "https://example.com/testing-graph"


def test_query_db(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"

    upload(
        sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    q = dedent("""
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            GRAPH ?g {
                ?c a skos:Concept .
            } 
        }""").replace("\n", "")

    result = runner.invoke(
        app,
        [
            "sparql",
            sparql_endpoint,
            q,
        ],
    )
    assert result.exit_code == 0


def test_query_file():
    # TODO: work out why this test fails but the direct call works
    # direct call:
    # kurra sparql /Users/nick/work/kurrawong/kurra/tests/test_sparql/language-test.ttl "PREFIX skos: <http://www.w3.org/2004/02/skos/core#> SELECT * WHERE {     ?c a skos:Concept .}"
    q = dedent("""
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            ?c a skos:Concept .
        }""").replace("\n", "")

    result = runner.invoke(
        app,
        ["sparql", str(LANG_TEST_VOC), q],
    )
    assert (
        "https://example.com/demo-vocabs/language-test/lang-and-no-lang"
        in result.output
    )


def test_select(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    result = runner.invoke(
        app,
        [
            "sparql",
            sparql_endpoint,
            "SELECT * WHERE { <https://example.com/demo-vocabs/language-test> ?p ?o }",
        ],
    )
    assert "https://example.com/demo-vocabs/lan" in result.output


def test_describe(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    result = runner.invoke(
        app,
        [
            "sparql",
            sparql_endpoint,
            "DESCRIBE <https://example.com/demo-vocabs/language-test>",
        ],
    )
    assert "Made in Nov 2024 just for testing" in result.output


def test_fuseki_sparql_drop(fuseki_container):
    result = runner.invoke(
        app,
        [
            "db",
            "sparql",
            "DROP ALL",
            f"http://localhost:{fuseki_container.get_exposed_port(3030)}",
            "-u",
            "admin",
            "-p",
            "admin",
        ],
    )
    # assert result.exit_code == 0  # TODO: work out why this isn't returning 0
    assert result.output == ""



import json
import pickle
from enum import Enum
from pathlib import Path
from typing import Union

import httpx
from rdflib import BNode, Dataset, Graph, Literal, Namespace, URIRef
from rdflib.plugins.parsers.notation3 import BadSyntax
from sparqlkit import (
    QuerySubType,
    SparqlStatementType,
    SparqlType,
    UpdateSubType,
    statement_type_from_string,
)

RDF_SUFFIX_MAP = {
    ".nt": "application/n-triples",
    ".nq": "application/n-quads",
    ".owl": "application/rdf+xml",
    ".ttl": "text/turtle",
    ".trig": "application/trig",
    ".json": "application/ld+json",
    ".jsonld": "application/ld+json",
    ".xml": "application/rdf+xml",
}

RDF_FILE_SUFFIXES = {
    "turtle": ".ttl",
    "longturtle": ".ttl",
    "xml": ".rdf",
    "n-triples": ".nt",
    "json-ld": ".jsonld",
    "owl": ".owl",
}

SYSTEM_GRAPH_IRI = URIRef("https://olis.dev/SystemGraph")

OLIS = Namespace("https://olis.dev/")


class RenderFormat(str, Enum):
    original = "original"
    json = "json"
    markdown = "markdown"


def guess_format_from_data(rdf: str) -> str | None:
    if rdf is not None:
        rdf = rdf.strip()
        if rdf.startswith("PREFIX") or rdf.startswith("@prefix"):
            return "text/turtle"
        elif rdf.startswith("{") or rdf.startswith("["):
            return "application/ld+json"
        elif rdf.startswith("<?xml") or rdf.startswith("<rdf"):
            return "application/rdf+xml"
        elif rdf.startswith("<http"):
            return "application/n-triples"
        else:
            return "application/n-triples"
    else:
        return None


def load_graph(graph_path_or_str: Union[Graph, Path, str], recursive=False) -> Graph:
    """
    Presents an RDFLib Graph object from a pre-existing Graph, a pickle file, an RDF file or directory of files or RDF
    data in a string
    """
    # Pre-existing Graph
    if isinstance(graph_path_or_str, Graph):
        return graph_path_or_str

    # Pickle file
    if isinstance(graph_path_or_str, Path):
        if graph_path_or_str.is_file():
            pkl_path = graph_path_or_str.with_suffix(".pkl")
            if pkl_path.is_file():
                return pickle.load(open(pkl_path, "rb"))

    # Serialized RDF file or dir of files
    if isinstance(graph_path_or_str, Path):
        if Path(graph_path_or_str).is_file():
            if str(graph_path_or_str).endswith(".trig") or str(
                graph_path_or_str
            ).endswith(".jsonld"):
                return Dataset().parse(str(graph_path_or_str))
            return Graph().parse(graph_path_or_str)
        elif Path(graph_path_or_str).is_dir():
            g = Graph()
            if recursive:
                gl = Path(graph_path_or_str).rglob("*.ttl")
            else:
                gl = Path(graph_path_or_str).glob("*.ttl")
            for f in gl:
                if f.is_file():
                    g.parse(f)
            return g

    # A remote file via HTTP
    elif isinstance(graph_path_or_str, str) and graph_path_or_str.startswith("http"):
        return Graph().parse(graph_path_or_str)

    # RDF data in a string
    else:
        return Graph().parse(
            data=graph_path_or_str,
            format=guess_format_from_data(graph_path_or_str),
        )


def render_sparql_result(
    r: dict | str | Graph, rf: RenderFormat = RenderFormat.markdown
) -> str:
    """Renders a SPARQL result in a given render format"""
    if rf == RenderFormat.original:
        return r

    elif rf == RenderFormat.json:
        if isinstance(r, dict):
            return json.dumps(r, indent=4)
        elif isinstance(r, str):
            return json.dumps(json.loads(r), indent=4)
        elif isinstance(r, Graph):
            return r.serialize(format="json-ld", indent=4)

    elif rf == RenderFormat.markdown:
        if isinstance(r, Graph):  # CONSTRUCT: RDF GRaph
            output = "```turtle\n" + r.serialize(format="longturtle") + "```\n"
        else:  # SELECT or ASK: Python dict or JSON

            def render_sparql_value(v: dict) -> str:
                # TODO: handle v["datatype"]
                if v is None:
                    return ""
                elif isinstance(v, URIRef) or isinstance(v, str):
                    return f"[{v.split('/')[-1].split('#')[-1]}]({v})"
                elif isinstance(v, Literal):
                    return v
                elif isinstance(v, BNode):
                    return f"BN: {v:>6}"
                elif v["type"] == "uri":
                    return f"[{v['value'].split('/')[-1].split('#')[-1]}]({v['value']})"
                elif v["type"] == "literal":
                    return v["value"]
                elif v["type"] == "bnode":
                    return f"BN: {v['value']:>6}"

            if isinstance(r, str):
                r = json.loads(r)

            output = ""
            header = ["", ""]
            body = []

            if r.get("head") is not None:
                # SELECT
                if r["head"].get("vars") is not None:
                    for col in r["head"]["vars"]:
                        header[0] += f"{col} | "
                        header[1] += f"--- | "
                    output = (
                        "| " + header[0].strip() + "\n| " + header[1].strip() + "\n"
                    )

            if r.get("results"):
                if r["results"].get("bindings"):
                    for row in r["results"]["bindings"]:
                        row_cols = []
                        for k in r["head"]["vars"]:
                            v = row.get(k)
                            if v is not None:
                                # ignore the k
                                row_cols.append(render_sparql_value(v))
                            else:
                                row_cols.append("")
                        body.append(" | ".join(row_cols))

                output += "\n| ".join(body) + " |\n"

            if r.get("boolean") is not None:
                output = str(bool(r.get("boolean")))

        return output


def make_httpx_client(
    sparql_username: str = None,
    sparql_password: str = None,
):
    auth = None
    if sparql_username:
        if sparql_password:
            auth = httpx.BasicAuth(sparql_username, sparql_password)
    return httpx.Client(auth=auth)


def convert_sparql_json_to_python(
    j: Union[str, bytes, httpx.Response], return_bindings_only=False
) -> {}:
    if type(j) == str:
        r = json.loads(j)
    elif type(j) == bytes:
        r = json.loads(j.decode())
    elif type(j) == httpx.Response:
        r = j.json()

    if r.get("results") is not None:  # SELECT
        for row in r["results"]["bindings"]:
            for k, v in row.items():
                if v["type"] == "literal":
                    if v.get("datatype") is not None:
                        row[k] = Literal(v["value"], datatype=v["datatype"]).toPython()
                    else:
                        row[k] = Literal(v["value"]).toPython()
                elif v["type"] == "uri":
                    row[k] = v["value"]
        if return_bindings_only:
            r = r["results"]["bindings"]
        return r
    elif r.get("boolean") is not None:  # ASK
        if return_bindings_only:
            return bool(r["boolean"])
        else:
            return r
    else:
        return r


def sparql_statement_return_type(
    query: str, statement: SparqlStatementType | None = None
) -> str:
    statement = _ensure_statement_type(query, statement)
    if is_construct_or_describe_query(query, statement):
        return "text/turtle"
    return "application/sparql-results+json"


def statement_type_for_query(query: str) -> SparqlStatementType:
    return statement_type_from_string(query)


def _ensure_statement_type(
    query: str, statement: SparqlStatementType | None = None
) -> SparqlStatementType:
    return statement if statement is not None else statement_type_for_query(query)


def is_construct_query(
    query: str, statement: SparqlStatementType | None = None
) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.QUERY
        and statement.subtype == QuerySubType.CONSTRUCT
    )


def is_describe_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.QUERY
        and statement.subtype == QuerySubType.DESCRIBE
    )


def is_select_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.QUERY and statement.subtype == QuerySubType.SELECT
    )


def is_ask_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.QUERY and statement.subtype == QuerySubType.ASK


def is_construct_or_describe_query(
    query: str, statement: SparqlStatementType | None = None
) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.QUERY and statement.subtype in {
        QuerySubType.CONSTRUCT,
        QuerySubType.DESCRIBE,
    }


def is_select_or_ask_query(
    query: str, statement: SparqlStatementType | None = None
) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.QUERY and statement.subtype in {
        QuerySubType.SELECT,
        QuerySubType.ASK,
    }


def is_update_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.UPDATE


def is_drop_update(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.UPDATE and statement.subtype == UpdateSubType.DROP
    )


def make_sparql_dataframe(sparql_result: dict):
    try:
        from pandas import DataFrame
    except ImportError:
        raise ValueError(
            'You selected the output format "dataframe" but the pandas Python package is not installed.'
        )

    if sparql_result.get("results") is not None:  # SELECT
        df = DataFrame(columns=sparql_result["head"]["vars"])
        for i, row in enumerate(sparql_result["results"]["bindings"]):
            new_row = {}
            for k, v in row.items():
                if v["type"] == "literal":
                    if v.get("datatype") is not None:
                        new_row[k] = Literal(
                            v["value"], datatype=v["datatype"]
                        ).toPython()
                    else:
                        new_row[k] = Literal(v["value"]).toPython()
                else:
                    new_row[k] = v["value"]
            df.loc[i] = new_row
        return df
    else:  # ASK
        df = DataFrame(columns=["boolean"])
        df.loc[0] = sparql_result["boolean"]

    return df


def add_namespaces_to_query_or_data(q: str, namespaces: dict):
    preamble = ""
    for k, v in namespaces.items():
        preamble += f"PREFIX {k}: <{v}>\n"
    preamble += "\n"
    return preamble + q


def make_httpx_client(
    sparql_username: str = None,
    sparql_password: str = None,
    timeout: int = 60,
):
    auth = None
    if sparql_username:
        if sparql_password:
            auth = httpx.BasicAuth(sparql_username, sparql_password)
    return httpx.Client(auth=auth, timeout=timeout)


def get_system_graph(
    system_graph_source: str | Path | Dataset | Graph = None,
    http_client: httpx.Client | None = None,
):
    """Returns a System Graph, graph and can accept many source options"""
    system_graph = Graph(identifier=SYSTEM_GRAPH_IRI)
    system_graph.bind("olis", OLIS)
    if system_graph_source is None:
        # no incoming System Graph
        pass
    elif isinstance(system_graph_source, Path):
        # we have a Graph or Dataset file, so read it
        if not system_graph_source.is_file():
            raise ValueError(
                f"system_graph_source must be an existing RDF file. Value supplied was {system_graph_source}"
            )

        if system_graph_source.suffix == ".trig":
            system_graph += (
                Dataset()
                .parse(system_graph_source, format="trig")
                .get_graph(SYSTEM_GRAPH_IRI)
            )
        else:
            system_graph += load_graph(system_graph_source)
    elif isinstance(system_graph_source, Graph):
        # we have a Graph, so assume it's a System Graph and load it
        system_graph += system_graph_source
    elif isinstance(system_graph_source, Dataset):
        # we have a Dataset object, so load its system Graph
        system_graph += system_graph_source.get_graph(str(SYSTEM_GRAPH_IRI))
    elif system_graph_source and system_graph_source.startswith("http"):
        # we have a remote SPARQL Endpoint, so read the System Graph
        # this is simplified GSP get()
        close_http_client = False
        if http_client is None:
            http_client = httpx.Client()
            close_http_client = True

        r = http_client.get(
            str(system_graph_source),
            params={"graph": SYSTEM_GRAPH_IRI},
            headers={"Accept": "text/turtle"},
        )

        if close_http_client:
            http_client.close()

        if r.is_success:
            system_graph += Graph().parse(data=r.text, format="turtle")
        else:
            return r.status_code
    elif system_graph_source and not system_graph_source.startswith("http"):
        system_graph += load_graph(system_graph_source)
    else:
        raise ValueError(
            "The parameter system_graph_source must be either None, a Path to an RDF Graph or Dataset serialised "
            "in Turtle or Trig, an RDFLib Graph object assumed to be a System Graph, an RDFLib Dataset object containing"
            "a System Graph or a string URL for a SPARQL Endpoint."
        )

    return system_graph


def put_system_graph(
    system_graph: Graph,
    system_graph_source: str | Path | Dataset | Graph = None,
    http_client: httpx.Client | None = None,
):
    if system_graph_source is None:
        return system_graph
    elif isinstance(system_graph_source, Path):
        if system_graph_source.suffix == ".trig":
            # TODO: deduplicate the Dataset parse in get_system_graph
            d = Dataset().parse(system_graph_source, format="trig")
            d.remove_graph(SYSTEM_GRAPH_IRI)
            d.add_graph(system_graph)
            d.serialize(destination=system_graph_source, format="trig")
        else:
            system_graph.serialize(destination=system_graph_source, format="longturtle")

        return None
    elif isinstance(system_graph_source, Graph):
        system_graph_source = system_graph
        return None
    elif isinstance(system_graph_source, Dataset):
        system_graph_source.remove_graph(SYSTEM_GRAPH_IRI)
        system_graph_source.add_graph(system_graph)
        return None
    elif system_graph_source and system_graph_source.startswith("http"):
        # this is simplified GSP put()
        close_http_client = False
        if http_client is None:
            http_client = httpx.Client()
            close_http_client = True

        r = http_client.put(
            system_graph_source,
            params={"graph": SYSTEM_GRAPH_IRI},
            headers={"Content-Type": "text/turtle"},
            content=system_graph.serialize(format="text/turtle"),
        )

        if close_http_client:
            http_client.close()

        if r.is_success:
            return None
        else:
            return r.status_code
    elif system_graph_source and not system_graph_source.startswith("http"):
        return system_graph
    else:
        return None



from pathlib import Path
from typing import Annotated

import typer

from kurra.cli.commands.db.fuseki import app
from kurra.cli.commands.sparql import sparql_command

app = typer.Typer(help="RDF Database commands")


@app.command(name="sparql", help="SPARQL query an RDF database")
def sparql_command3(
    path_or_url: Path = typer.Argument(
        ..., help="Repository SPARQL Endpoint URL. E.g. http://localhost:3030/ds"
    ),
    q: str = typer.Argument(..., help="The SPARQL query to sent to the database"),
    response_format: Annotated[
        str,
        typer.Option(
            "--response-format",
            "-f",
            help="The response format of the SPARQL query",
        ),
    ] = "table",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    sparql_command(path_or_url, q, response_format, username, password, timeout)



from pathlib import Path

import pytest
from rdflib import URIRef
from rdflib.namespace import RDF
from typer.testing import CliRunner

from kurra.cli import app
from kurra.db.gsp import get, put, upload
from kurra.sparql import query

runner = CliRunner()


def test_exists(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    upload(SPARQL_ENDPOINT, f, TESTING_GRAPH, http_client=http_client)

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "exists",
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )
    assert "True" in result.output

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "exists",
            "-g",
            "http://nothing.com",
            SPARQL_ENDPOINT,
        ],
    )
    assert "False" in result.output

    # default graph
    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "exists",
            SPARQL_ENDPOINT,
        ],
    )
    assert "True" in result.output


def test_get(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    upload(SPARQL_ENDPOINT, f, TESTING_GRAPH, http_client=http_client)

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "get",
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )
    assert "PREFIX bibo:" in result.output

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "get",
            "-g",
            "http://nothing.com",
            SPARQL_ENDPOINT,
        ],
    )
    assert "Graph not found" in result.output

    # default graph
    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "get",
            SPARQL_ENDPOINT,
        ],
    )
    assert "No content" in result.output


def test_put(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    runner.invoke(
        app,
        [
            "db",
            "gsp",
            "put",
            str(f),
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert (
        URIRef("http://base/#service_tdb_all"),
        RDF.type,
        URIRef("http://jena.apache.org/fuseki#Service"),
    ) in r

    # delete(SPARQL_ENDPOINT, "all", http_client=http_client)

    r = get(SPARQL_ENDPOINT, http_client=http_client)
    assert len(r) == 0

    runner.invoke(
        app,
        [
            "db",
            "gsp",
            "put",
            str(Path(__file__).parent.parent.resolve().parent / "rdf" / "rdf_1.ttl"),
            "-g",
            "default",
            # no -g so default
            SPARQL_ENDPOINT,
        ],
    )
    r = get(SPARQL_ENDPOINT, http_client=http_client)
    assert (
        URIRef("http://example.com/a"),
        URIRef("http://example.com/b"),
        URIRef("http://example.com/c"),
    ) in r


def test_post(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    put(SPARQL_ENDPOINT, f, TESTING_GRAPH, http_client=http_client)

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert len(r) == 142

    runner.invoke(
        app,
        [
            "db",
            "gsp",
            "post",
            str(Path(__file__).parent.parent.resolve().parent / "rdf" / "rdf_1.ttl"),
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert len(r) == 143


def test_delete(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    put(SPARQL_ENDPOINT, f, TESTING_GRAPH, http_client=http_client)

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert len(r) == 142

    runner.invoke(
        app,
        [
            "db",
            "gsp",
            "delete",
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert r == 404


def test_clear(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    put(SPARQL_ENDPOINT, f, TESTING_GRAPH, http_client=http_client)

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert len(r) == 142

    runner.invoke(
        app,
        [
            "db",
            "gsp",
            "delete",
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )

    r = get(SPARQL_ENDPOINT, TESTING_GRAPH, http_client=http_client)
    assert r == 404


@pytest.mark.xfail  # This test fails with the testcontainer Fuseki image but works on 'rea' Fuseki installations
def test_upload_file(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    f = Path(__file__).parent / "config.ttl"

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "upload",
            str(f),
            SPARQL_ENDPOINT,
        ],
    )
    assert result.exit_code == 0

    q = """
        SELECT (COUNT(?s) AS ?count)
        WHERE {
            ?s ?p ?o
        }
        """
    r = query(SPARQL_ENDPOINT, q, return_format="python", return_bindings_only=True)
    assert r[0]["count"] == 142


def test_upload_file_with_graph_id(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    f = Path(__file__).parent / "config.ttl"

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "upload",
            str(f),
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )
    assert result.exit_code == 0

    q = """
        SELECT (COUNT(?s) AS ?count)
        WHERE {
            ?s ?p ?o
        }
        """
    r = query(SPARQL_ENDPOINT, q, return_format="python", return_bindings_only=True)
    assert r[0]["count"] == 142


def test_upload_file_with_graph_id_file(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "file"
    f = Path(__file__).parent / "config.ttl"

    result = runner.invoke(
        app,
        [
            "db",
            "gsp",
            "upload",
            str(f),
            "-g",
            TESTING_GRAPH,
            SPARQL_ENDPOINT,
        ],
    )
    assert result.exit_code == 0

    q = """
        SELECT DISTINCT ?g (COUNT(?s) AS ?count)
        WHERE {
        GRAPH ?g {
                ?s ?p ?o
            }
        }
        GROUP BY ?g
        """
    r = query(SPARQL_ENDPOINT, q, return_format="python", return_bindings_only=True)
    assert r[0]["count"] == 142
    assert r[0]["g"] == "urn:file:config.ttl"



import httpx

from kurra.db.gsp import upload
from kurra.sparql import query


def test_query(fuseki_container, http_client):
    with httpx.Client() as client:
        sparql_endpoint = (
            f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
        )
        testing_graph = "https://example.com/testing-graph"

        data = """
                PREFIX ex: <http://example.com/>
                
                ex:a ex:b ex:c .
                ex:a2 ex:b2 ex:c2 .
                """

        upload(sparql_endpoint, data, testing_graph, False, http_client=http_client)

        q = """
            SELECT (COUNT(*) AS ?count) 
            WHERE {
              GRAPH <XXX> {
                ?s ?p ?o
              }
            }        
            """.replace("XXX", testing_graph)

        r = query(
            sparql_endpoint,
            q,
            http_client=http_client,
            return_format="python",
            return_bindings_only=True,
        )

        assert r[0]["count"] == 2

        q = "DROP GRAPH <XXX>".replace("XXX", testing_graph)

        r = query(sparql_endpoint, q, http_client=http_client)

        q = """
            SELECT (COUNT(*) AS ?count) 
            WHERE {
              GRAPH <XXX> {
                ?s ?p ?o
              }
            }        
            """.replace("XXX", testing_graph)

        r = query(
            sparql_endpoint,
            q,
            http_client=http_client,
            return_format="python",
            return_bindings_only=True,
        )

        assert r[0]["count"] == 0



from typing import Literal as LiteralType

import httpx

from kurra.utils import (
    add_namespaces_to_query_or_data,
    convert_sparql_json_to_python,
    is_construct_or_describe_query,
    is_select_or_ask_query,
    is_update_query,
    make_sparql_dataframe,
    sparql_statement_return_type,
    statement_type_for_query,
)


def query(
    sparql_endpoint: str,
    q: str,
    namespaces: dict[str, str] | None = None,
    http_client: httpx.Client = None,
    return_format: LiteralType["original", "python", "dataframe"] = "original",
    return_bindings_only: bool = False,
):
    """Pose a SPARQL query to a SPARQL Endpoint"""
    if sparql_endpoint is None:
        raise ValueError("You must supply a sparql_endpoint")

    if q is None:
        raise ValueError("You must supply a query")

    if return_format not in ["original", "python", "dataframe"]:
        raise ValueError(
            f"return_format {return_format} must be either 'original', 'python' or 'dataframe'"
        )

    if namespaces is not None:
        q = add_namespaces_to_query_or_data(q, namespaces)

    if http_client is None:
        http_client = httpx.Client()

    statement = statement_type_for_query(q)

    if return_format == "dataframe":
        if not is_select_or_ask_query(q, statement):
            raise ValueError(
                'Only SELECT and ASK queries can have return_format set to "dataframe"'
            )

        try:
            from pandas import DataFrame
        except ImportError:
            raise ValueError(
                'You selected the output format "dataframe" but the pandas Python package is not installed.'
            )

    if is_update_query(q, statement):
        headers = {"Content-Type": "application/sparql-update"}
    else:
        headers = {"Content-Type": "application/sparql-query"}

    headers["Accept"] = sparql_statement_return_type(q, statement)

    r = http_client.post(
        sparql_endpoint,
        headers=headers,
        content=q,
    )

    status_code = r.status_code

    # in case the endpoint doesn't allow POST
    if status_code == 405 or status_code == 422:
        r = http_client.get(
            sparql_endpoint,
            headers=headers,
            params={"query": q},
        )

        status_code = r.status_code

    if status_code != 200 and status_code != 201 and status_code != 204:
        raise RuntimeError(f"ERROR {status_code}: {r.text}")

    if status_code == 204:
        return ""

    if is_construct_or_describe_query(q, statement):
        return r.text

    if return_format == "python":
        return convert_sparql_json_to_python(r, return_bindings_only)

    elif return_format == "dataframe":
        return make_sparql_dataframe(r.json())

    # original format - JSON
    return r.text



from pathlib import Path
from typing import Annotated

import httpx
import typer

from kurra.cli.console import console
from kurra.db.fuseki import FusekiError, create, delete, describe

app = typer.Typer(help="Fuseki database commands")

dataset_type_options = ["mem", "tdb", "tdb1", "tdb2"]


@app.command(name="ping", help="Check if the server is alive")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="server", help="Get basic server info")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="stats", help="Request statistics for all datasets")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="backup", help="Ask the server to create a backup")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="backups_list", help="List all existing backups")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="sleep", help="Tell the server to sleep")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="tasks", help="List running tasks")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="metrics", help="Get server metrics")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(name="describe", help="Get the list of datasets or describe one")
def describe_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(describe(fuseki_url, http_client=http_client))
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to describe datasets at {fuseki_url}."
            )
            raise err


@app.command(
    name="create",
    help="Create a new dataset",
)
def create_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    dataset_name: str | None = typer.Argument(None, help="repository name"),
    dataset_type: str = typer.Option(
        "tdb2", help=f"dataset type. Options: {dataset_type_options}"
    ),
    config: Path | None = typer.Option(None, help="assembler file"),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    if dataset_name and config:
        raise typer.BadParameter("Only dataset name or --config is allowed, not both.")

    if dataset_name and dataset_type not in dataset_type_options:
        raise typer.BadParameter(
            f"Invalid dataset type '{dataset_type}'. Options: {dataset_type_options}"
        )

    if dataset_name:
        with httpx.Client(auth=auth, timeout=timeout) as http_client:
            try:
                console.print(
                    create(
                        fuseki_url, dataset_name, dataset_type, http_client=http_client
                    )
                )
            except Exception as err:
                console.print(
                    f"[bold red]ERROR[/bold red] Failed to create dataset {dataset_name} at {fuseki_url}."
                )
                raise err
    else:
        if config is None:
            raise typer.BadParameter(
                "Either dataset name or assembler config file must be provided."
            )
        with httpx.Client(auth=auth, timeout=timeout) as http_client:
            try:
                console.print(create(fuseki_url, config, http_client=http_client))
            except Exception as err:
                console.print(
                    f"[bold red]ERROR[/bold red] Failed to create dataset {dataset_name} at {fuseki_url}."
                )
                raise err


@app.command(name="delete", help="Delete a dataset")
def delete_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    dataset_name: str = typer.Argument(..., help="The name of the dataset to delete."),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(delete(fuseki_url, dataset_name, http_client=http_client))
        except FusekiError as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to delete dataset {dataset_name} at {fuseki_url}."
            )
            raise err



import shutil
from pathlib import Path
from pickle import dump, load

import pytest
from rdflib import Dataset, URIRef
from rdflib.namespace import RDF, SH

from kurra.shacl import (
    check_validator_known,
    list_local_validators,
    sync_validators,
    validate,
)
from kurra.sparql import query
from kurra.utils import load_graph

SHACL_TEST_DIR = Path(__file__).parent.resolve()


def test_validate_simple():
    shacl_graph = load_graph(SHACL_TEST_DIR / "validator-vocpub-410.ttl")

    data_file = SHACL_TEST_DIR / "vocab-valid.ttl"
    valid, g, txt = validate(data_file, shacl_graph)
    assert valid

    data_file2 = SHACL_TEST_DIR / "vocab-invalid.ttl"
    valid2, g2, txt2 = validate(data_file2, shacl_graph)
    assert not valid2

    data_file3 = SHACL_TEST_DIR / "vocab-invalid2.ttl"
    valid3, g3, txt3 = validate(data_file3, shacl_graph)
    assert not valid3


def test_validate_multiple_data_files():
    data_files = [
        Path(__file__).parent / "vocab-invalid.ttl",
        Path(__file__).parent / "vocab-invalid-additions.ttl",
    ]
    v = validate(data_files, "https://linked.data.gov.au/def/vocpub/validator")
    assert v[0]


def test_sync_validators():
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"

    # get number of SemBack validators
    q = """
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX schema: <https://schema.org/>
        
        SELECT (COUNT(*) AS ?count)
        WHERE {
          <https://data.kurrawong.ai/sb/validators>
                schema:hasPart ?o ;
          .
          
          ?o a owl:Ontology .
        }
        """
    num_remote_validators = query(
        "https://fuseki.dev.kurrawong.ai/semback/sparql",
        q,
        return_format="python",
        return_bindings_only=True,
    )[0]["count"]

    if Path.is_dir(kurra_cache):
        shutil.rmtree(kurra_cache)

    known_validators = sync_validators()

    assert len(known_validators) == num_remote_validators

    d = load(open(validators_cache, "rb"))
    d: Dataset
    d.remove_graph(d.get_graph(URIRef("https://prez.dev/manifest-validator")))
    with open(validators_cache, "wb") as f:
        dump(d, f)

    assert len(list_local_validators().keys()) == (num_remote_validators - 1)

    known_validators = sync_validators()

    assert len(known_validators) == num_remote_validators


def test_list_local_validators():
    pm_cache = Path().home() / ".pm"

    if Path.is_dir(pm_cache):
        shutil.rmtree(pm_cache)

    sync_validators()

    assert len(list_local_validators().keys()) == 12


def test_validate_by_id():
    """Awaiting sync_validators()"""
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"

    sync_validators()

    valid, g, txt = validate(SHACL_TEST_DIR / "vocab-valid.ttl", 9)
    assert (
        len(list(g.subjects(predicate=RDF.type, object=SH.ValidationResult))) == 1
    )  # Warning

    valid, g, txt = validate(SHACL_TEST_DIR / "vocab-invalid.ttl", 9)
    assert len(list(g.subjects(predicate=RDF.type, object=SH.ValidationResult))) == 6


def test_check_validator_known():
    assert check_validator_known("https://linked.data.gov.au/def/vocpub/validator")
    assert not check_validator_known("https://linked.data.gov.au/def/vocpub/validatorx")



# Graph Store Protocol

from pathlib import Path
from typing import Literal as LiteralType
from typing import Union

import httpx
from rdflib import Graph

from kurra.utils import RDF_SUFFIX_MAP, load_graph


def exists(
    sparql_endpoint: str, graph_iri: str, http_client: httpx.Client | None = None
) -> bool:
    """Returns True if a graph with the given graph_iri exists at the SPARQL Endpoint or else False"""
    if not sparql_endpoint.startswith("http"):
        raise ValueError(f"SPARQL Endpoint given does not start with 'http'")

    if not graph_iri:
        raise ValueError("You must supply a graph IRI")

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.head(
        sparql_endpoint,
        params={"graph": graph_iri if graph_iri is not None else "default"},
    )

    if close_http_client:
        http_client.close()

    return r.is_success


def get(
    sparql_endpoint: str,
    graph_iri: str = "default",
    accept_type="text/turtle",
    return_format: LiteralType["original", "python"] = "python",
    http_client: httpx.Client | None = None,
) -> Union[Graph, int]:
    """Graph Store Protocol's HTTP GET: https://www.w3.org/TR/sparql12-graph-store-protocol/#http-get

    Returns the content of the graph identified by graph_id in the target SPARQL Endpoint.

    Args:
        sparql_endpoint: The SPARQL Endpoint URL to use
        graph_iri: The IRI of the graph to retrieve
        accept_type: The RDF format to request from the server and to return if return_format is set to 'original'
        return_format: The return format to use, 'python' - RDFLib's Graph - or 'original' - an RDF string value in the format of accept_type
        http_client: An HTTP client to use. Created internally if not supplied

    Returns:
          An RDF result as either an RDFLib Graph object or a string object containing RDF in the accept_type
          format. If a graph, the graph identifier will be the graph_iri or a Blank Node if None/default
    """
    if not sparql_endpoint.startswith("http"):
        raise ValueError(f"SPARQL Endpoint given does not start with 'http'")

    if accept_type not in RDF_SUFFIX_MAP.values():
        raise ValueError(
            f"Media Type requested not available. Allow types are {', '.join(RDF_SUFFIX_MAP.values())}"
        )

    if return_format not in ["original", "python"]:
        raise ValueError(
            "Return format must be either 'python' (default) or 'original'"
        )

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.get(
        sparql_endpoint,
        params={"graph": graph_iri if graph_iri is not None else "default"},
        headers={"Accept": accept_type},
    )

    if close_http_client:
        http_client.close()

    if r.is_success:
        if return_format == "original":
            return r.text
        else:
            if graph_iri is not None and graph_iri != "default":
                return Graph(identifier=graph_iri).parse(
                    data=r.text, format=accept_type
                )
            else:
                return Graph().parse(data=r.text, format=accept_type)
    else:
        return r.status_code


def put(
    sparql_endpoint: str,
    file_or_str_or_graph: Union[Path, str, Graph],
    graph_iri: str = "default",
    content_type="text/turtle",
    http_client: httpx.Client | None = None,
) -> Union[Graph, int]:
    """Graph Store Protocol's HTTP PUT: https://www.w3.org/TR/sparql12-graph-store-protocol/#http-put

    Inserts the RDF content supplied into a graph identified by graph_id or the default graph.

    Will replace existing content."""
    if not sparql_endpoint.startswith("http"):
        raise ValueError(f"SPARQL Endpoint given does not start with 'http'")

    if content_type not in RDF_SUFFIX_MAP.values():
        raise ValueError(
            f"Media Type {content_type} requested not available. Allowed types are {', '.join(RDF_SUFFIX_MAP.values())}"
        )

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True
    r = http_client.put(
        sparql_endpoint,
        params={"graph": graph_iri if graph_iri is not None else "default"},
        headers={"Content-Type": content_type},
        content=load_graph(file_or_str_or_graph).serialize(format=content_type),
    )

    if close_http_client:
        http_client.close()

    if r.is_success:
        return True
    else:
        return r.status_code


def post(
    sparql_endpoint: str,
    file_or_str_or_graph: Union[Path, str, Graph],
    graph_iri: str = "default",
    content_type="text/turtle",
    http_client: httpx.Client | None = None,
) -> Union[Graph, int]:
    """Graph Store Protocol's HTTP POST: https://www.w3.org/TR/sparql12-graph-store-protocol/#http-post

    Inserts the RDF content supplied into a graph identified by graph_id or the default graph.

    Will add to existing content."""
    if not sparql_endpoint.startswith("http"):
        raise ValueError(f"SPARQL Endpoint given does not start with 'http'")

    if content_type not in RDF_SUFFIX_MAP.values():
        raise ValueError(
            f"Media Type requested not available. Allow types are {', '.join(RDF_SUFFIX_MAP.values())}"
        )

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.post(
        sparql_endpoint,
        params={"graph": graph_iri if graph_iri is not None else "default"},
        headers={
            "Content-Type": content_type,
        },
        content=load_graph(file_or_str_or_graph).serialize(format=content_type),
    )

    if close_http_client:
        http_client.close()

    if r.is_success:
        return True
    else:
        return r.status_code


def delete(
    sparql_endpoint: str,
    graph_iri: str = "default",
    http_client: httpx.Client | None = None,
) -> Union[Graph, int]:
    """Graph Store Protocol's HTTP DELETE: https://www.w3.org/TR/sparql12-graph-store-protocol/#http-delete

    Deletes the graph identified by graph_id or the default graph."""
    if not sparql_endpoint.startswith("http"):
        raise ValueError(f"SPARQL Endpoint given does not start with 'http'")

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.delete(
        sparql_endpoint,
        params={"graph": graph_iri if graph_iri is not None else "default"},
    )

    if close_http_client:
        http_client.close()

    if r.is_success:
        return True
    else:
        return r.status_code


def clear(
    sparql_endpoint: str, graph_iri: str, http_client: httpx.Client | None = None
):
    """SPARQL Update Clear function: https://www.w3.org/TR/sparql12-update/#clear

    Clears - remove all triples from - an identified graph or from all graphs if "all" is given as the graph_id.

    This is an alias of delete()
    """
    delete(sparql_endpoint, graph_iri, http_client)


def upload(
    sparql_endpoint: str,
    file_or_str_or_graph: Union[Path, str, Graph],
    graph_id: str | None = None,
    append: bool = False,
    content_type: str = "text/turtle",
    http_client: httpx.Client | None = None,
) -> Union[bool, int]:
    """This function uploads a file to a SPARQL Endpoint using the Graph Store Protocol.

    It will upload it into a graph identified by graph_id (an IRI or Blank Node). If no graph_id is given, it will be
    uploaded into the default graph.

    By default, it will replace all content in the Named Graph or default graph. If append is set to True, it will
    add it to existing content in the graph_id Named Graph.

    This function is an alias of put() (append=False) and post() (append=True)."""

    if append:
        return post(
            sparql_endpoint, file_or_str_or_graph, graph_id, content_type, http_client
        )
    else:
        return put(
            sparql_endpoint, file_or_str_or_graph, graph_id, content_type, http_client
        )



from kurra.cli.commands.db import app
from kurra.cli.commands.db.fuseki import app as fuseki_app
from kurra.cli.commands.db.gsp import app as gsp_app
from kurra.cli.commands.db.olis import app as olis_app

app.add_typer(fuseki_app, name="fuseki")
app.add_typer(gsp_app, name="gsp")
app.add_typer(olis_app, name="olis")



from pathlib import Path
from typing import Annotated

import httpx
import typer
from rich.progress import track

from kurra.cli.console import console
from kurra.db.gsp import clear, delete, exists, get, post, put, upload
from kurra.utils import RDF_SUFFIX_MAP

app = typer.Typer(help="Graph Store Protocol commands")


@app.command(name="exists", help="Checks to see if a graph exists within a database")
def exists_command(
    sparql_endpoint_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    graph_identifier: Annotated[
        str,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = "default",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(
                exists(sparql_endpoint_url, graph_identifier, http_client=http_client)
            )
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run clear command with '{graph_identifier}' at {sparql_endpoint_url}."
            )
            raise err


@app.command(name="get", help="Gets the content of a database graph")
def get_command(
    sparql_endpoint_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    graph_identifier: Annotated[
        str,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = "default",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            r = get(sparql_endpoint_url, graph_identifier, http_client=http_client)
            if r == 404:
                console.print("Graph not found")
            else:
                rdf = r.serialize(format="longturtle")
                if rdf == "":
                    console.print("No content")
                else:
                    console.print(rdf)

        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run the get command with '{graph_identifier}' at {sparql_endpoint_url}: {err.message}."
            )
            raise err


@app.command(name="put", help="Load content into a database graph")
def put_command(
    path: Path = typer.Argument(
        ..., help="The path of a file or directory of files to be uploaded."
    ),
    sparql_endpoint_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    graph_identifier: Annotated[
        str,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = "default",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(
                put(
                    sparql_endpoint_url, path, graph_identifier, http_client=http_client
                )
            )
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run clear command with '{graph_identifier}' at {sparql_endpoint_url}."
            )
            raise err


@app.command(name="post", help="Add content to a database graph")
def post_command(
    path: Path = typer.Argument(
        ..., help="The path of a file or directory of files to be uploaded."
    ),
    sparql_endpoint_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    graph_identifier: Annotated[
        str,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = "default",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(
                post(
                    sparql_endpoint_url, path, graph_identifier, http_client=http_client
                )
            )
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run clear command with '{graph_identifier}' at {sparql_endpoint_url}."
            )
            raise err


@app.command(name="delete", help="Deletes the content of a database graph")
def delete_command(
    sparql_endpoint_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    graph_identifier: Annotated[
        str,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = "default",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            console.print(
                delete(sparql_endpoint_url, graph_identifier, http_client=http_client)
            )
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run clear command with '{graph_identifier}' at {sparql_endpoint_url}."
            )
            raise err


@app.command(name="clear", help="Clears a database graph")
def clear_command(
    sparql_endpoint_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    graph_identifier: Annotated[
        str,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = "default",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        try:
            clear(sparql_endpoint_url, graph_identifier, http_client=http_client)
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run clear command with '{graph_identifier}' at {sparql_endpoint_url}."
            )
            raise err


@app.command(name="upload", help="Upload file(s) to a database")
def upload_command(
    path: Path = typer.Argument(
        ..., help="The path of a file or directory of files to be uploaded."
    ),
    sparql_endpoint: str = typer.Argument(
        ..., help="SPARQL Endpoint URL. E.g. http://localhost:3030/ds"
    ),
    graph_identifier: Annotated[
        str | None,
        typer.Option(
            "--graph",
            "-g",
            help='ID - IRI or URN - of the graph to upload into. If not set, the default graph is targeted. If set to the string "file", the URN urn:file:FILE_NAME will be used per file',
        ),
    ] = None,
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
    disable_ssl_verification: Annotated[
        bool,
        typer.Option(
            "--disable-ssl-verification", "-k", help="Disable SSL verification."
        ),
    ] = False,
    host_header: Annotated[
        str | None, typer.Option("--host-header", "-e", help="Override the Host header")
    ] = None,
) -> None:
    """Upload a file or a directory of files with an RDF file extension.

    File extensions: [.nt, .nq, .ttl, .trig, .json, .jsonld, .xml]

    Files are uploaded into their own named graph in the format:
    <urn:file:{file.name}>
    E.g. <urn:file:example.ttl>
    """
    files = []

    if path.is_file():
        files.append(path)
    else:
        files += path.glob("**/*")

    auth = (
        (username, password) if username is not None and password is not None else None
    )

    files = list(filter(lambda f: f.suffix in RDF_SUFFIX_MAP.keys(), files))

    with httpx.Client(
        auth=auth,
        timeout=timeout,
        headers={"Host": host_header} if host_header is not None else {},
        verify=False if disable_ssl_verification else True,
    ) as http_client:
        for file in track(files, description=f"Uploading {len(files)} files..."):
            try:
                if graph_identifier == "file":
                    upload(
                        sparql_endpoint,
                        file,
                        f"urn:file:{file.name}",
                        http_client=http_client,
                    )
                else:
                    upload(
                        sparql_endpoint,
                        file,
                        graph_identifier if graph_identifier is not None else "default",
                        http_client=http_client,
                    )  # str and None handled by upload()
            except Exception as err:
                console.print(
                    f"[bold red]ERROR[/bold red] Failed to upload file {file}."
                )
                raise err



import json
from pathlib import Path
from typing import Literal

import httpx
from rdflib import Dataset, Graph

from kurra.db.sparql import query as db_query
from kurra.utils import (
    add_namespaces_to_query_or_data,
    convert_sparql_json_to_python,
    is_construct_or_describe_query,
    is_drop_update,
    is_select_or_ask_query,
    is_update_query,
    load_graph,
    make_sparql_dataframe,
    statement_type_for_query,
)


def query(
    p: Path | str | Graph | Dataset,
    q: str,
    namespaces: dict[str, str] | None = None,
    http_client: httpx.Client = None,
    return_format: Literal["original", "python", "dataframe"] = "original",
    return_bindings_only: bool = False,
):
    """Pose a SPARQL query to a file, and RDF Graph or a SPARQL Endpoint"""
    if p is None:
        raise ValueError(
            "You must supply a Path, string (of data or a URL), Graph or a Dataset to query for variable p"
        )

    if q is None:
        raise ValueError("You must supply a query")

    if return_format not in ["original", "python", "dataframe"]:
        raise ValueError(
            f"return_format {return_format} must be either 'original', 'python' or 'dataframe'"
        )

    if namespaces is not None:
        q = add_namespaces_to_query_or_data(q, namespaces)

    if http_client is None:
        http_client = httpx.Client()

    statement = statement_type_for_query(q)

    if return_format == "dataframe":
        if not is_select_or_ask_query(q, statement):
            raise ValueError(
                'Only SELECT and ASK queries can have return_format set to "dataframe"'
            )

        try:
            from pandas import DataFrame
        except ImportError:
            raise ValueError(
                'You selected the output format "dataframe" but the pandas Python package is not installed.'
            )

    if is_construct_or_describe_query(q, statement):
        s = None
        f = None
        if str(p).startswith("http"):
            r = db_query(p, q, namespaces, http_client, "original", False)
            s = load_graph(r)

        else:  # (isinstance(p, str) and not p.startswith("http")) or isinstance(p, Path):
            f = load_graph(p).query(q)

        if return_format == "dataframe":
            raise ValueError(
                "DataFrames cannot be returned for CONSTRUCT or DESCRIBE queries"
            )
        elif return_format == "python":
            return s if s is not None else f.graph
        else:
            return (
                s.serialize(format="longturtle")
                if s is not None
                else f.graph.serialize(format="longturtle")
            )

    elif is_update_query(q, statement):
        if str(p).startswith("http"):
            close_http_client = False
            if http_client is None:
                http_client = httpx.Client()
                close_http_client = True

            r = db_query(p, q, namespaces, http_client, return_format, False)

            if close_http_client:
                http_client.close()

            if r == "" or r is None:
                return ""

        if is_drop_update(q, statement) and not isinstance(p, Dataset):
            raise NotImplementedError(
                f"DROP commands cannot be applied to Graphs or files or triples data, only Datasets or RDF DBs. You specified {p}"
            )
        elif isinstance(p, (Graph, str, Path)):
            g = load_graph(p)
            g.update(q)
            return g
        else:
            raise NotImplementedError(
                "Update SPARQL commands on Datasets are not yet supported"
            )

    else:  # SELECT or ASK
        r = None
        if str(p).startswith("http"):
            close_http_client = False
            if http_client is None:
                http_client = httpx.Client()
                close_http_client = True

            r = db_query(
                p, q, namespaces, http_client, return_format, return_bindings_only
            )

            if close_http_client:
                http_client.close()

        if r is not None:  # we have a result from the DB query to return
            return r
        else:  # querying a file or string RDF data
            r = load_graph(p).query(q).serialize(format="json")

            if return_format == "dataframe":
                return make_sparql_dataframe(json.loads(r))
            elif return_format == "python":
                return convert_sparql_json_to_python(r, return_bindings_only)
            else:
                return r.decode()
